/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "approximation/approximationUtils.h"

int main(int argc,char **argv) {
	if(argc < 7) {
		cout << "Usage: approximation <.mbin file> <.stac file> -k <k value> -m <m value> -v (optional, verbose)\n";
		return 1;
	}
	int K = 0, M = 0, V = 0;
	int argi = 3;
	while(argi < argc) {
		if(strcmp(argv[argi], "-k") == 0)
			K = atoi(argv[argi + 1]);
		if(strcmp(argv[argi], "-m") == 0)
			M = atoi(argv[argi + 1]);
		if(strcmp(argv[argi], "-v") == 0)
			V = 1;
		argi++;
	}
	if((K == 0) || (M == 0)) {
		cout << "Usage: approximation <.mbin file> <.stac file> -k <k value> -m <m value> -v (optional, verbose)\n";
		return 1;
	}
	ifstream finmbin(argv[1], ios::in|ios::binary);
	ofstream foutstac(argv[2], ios::out);
	StdVectorFst *ifst = new StdVectorFst;
	mbintofst(ifst, finmbin);
	int numnodes = getNumNodesFst(ifst);
	if(numnodes < 2) {
		cout << "Error: Fst is too small!\n";
		exit(1);
	}
	
	//obtain adjlist of vertices
	map <int, set<int> > adjlist;		//this is on directed edges
	for (StateIterator< Fst<StdArc> > siter(*ifst);!siter.Done(); siter.Next()) {
		int from = siter.Value();
		adjlist[from];
		for(ArcIterator< Fst<StdArc> > aiter(*ifst, from);!aiter.Done(); aiter.Next()) {
			StdArc arc = aiter.Value();
			int to = arc.nextstate;
			adjlist[from].insert(to);
		}
	}
	int nodedismin[numnodes*numnodes];
	floydstuff(ifst, nodedismin);	//if nodedismin[i][j] is INFIN, then there is no path from i to j
	
	map < Graph, block> allblocks;		//a list of all current blocks, each keyed by its origfstedges set
	vector <blockskel> allblockskels;		
	//start with each edge of fst as a block; then give a set of nodes to function, and it reconstructs induced subfst, enumerates its paths and returns block
	if(V) cout << "Computing base blocks\n";
	for(map <int, set<int> >::iterator iter = adjlist.begin(); iter != adjlist.end(); iter++) {		
		for(set<int>::iterator siter = (iter->second).begin(); siter != (iter->second).end(); siter++) {
			blockskel thisblkskel;
			thisblkskel.minnode = iter->first;
			thisblkskel.maxnode = *siter;
			(thisblkskel.origfstnodes).insert(iter->first);
			(thisblkskel.origfstnodes).insert(*siter);
			thisblkskel.origfstedges.insert(make_pair(iter->first, *siter));
			allblockskels.push_back(thisblkskel);
			
			block thisblk = getBlockData(ifst, thisblkskel, numnodes, K, 1);
			allblocks[thisblkskel.origfstedges] = thisblk;
		}
	}
	double overallprob = edgePushDynProb(&allblockskels, &allblocks);
	if(V) cout << "Overall probability of base blocks: " << overallprob << endl;
	
	//stacfile schema:
	//<M> <K> <number of edges in fst> <numnodes in fst>\n
	//<final numblks> <overall prob of final blocks set>\n
	foutstac << M << " " << K << " " << allblocks.size() << " " << numnodes << endl;

	//memoize all candidates; when starting, the candidate set is empty; after first iteration, we get a huge set
	//we pick the best candidate and make it a blk in the subseq iteration, removing it from the candidates set
	//no, we ascertain what candidates overlap partially with the new block's skel, and remove them (do not remove if the candidate supersets the new blk)
	//next, we ascertain a set of possible cand addns -> based on ngbrs of the new block's skel; add the new set of cnadts to the memoize candt set
	//we pick the best and repeat this process till M
	map < Graph, blockskel > memocandskels;	//memo of candidate skels 
	//check numbaseblks and M, if more, we need to get the first set of candidates
	if(V) cout << "Number of base blocks: " << allblocks.size() << " and m " << M << endl;	
	if(allblocks.size() <= M) {
		if(V) cout << "Done. We have final blocks\n";
	}
	else {	//get first canddt set and memoize; then start iterations
		if(V) cout << "Need to merge blocks\n";
		map < Graph, blockskel > memocandskels;
		map < Graph, double > memocandskelprobs;
		
		map < Graph, blockskel > candmergeskels;
		map < pair < Graph, Graph >, int > startpairskels;	//for each cand, we start with a pair of legal skels
		
		for(vector <blockskel>::iterator blkiter = allblockskels.begin(); blkiter != allblockskels.end(); blkiter++) {
			vector <blockskel> ngbrskels = getNgbrSkels(*blkiter, &allblockskels);		//get list of nghbrng blkskels
			if(V) cout << "Considering neighbors of block (min,mix) " << blkiter->minnode << ", " << blkiter->maxnode ;		
			if(V) cout << " (" << ngbrskels.size() << " neighbors)\n";
			for(vector <blockskel>::iterator ngbriter = ngbrskels.begin(); ngbriter != ngbrskels.end(); ngbriter++) {	//for each ngbr we do for loop
				if(ngbriter->origfstedges == blkiter->origfstedges) //cannot merge a legal blk with itself
					continue;
				//an optmzn - no need to consider a start-pair that was seen (in rev) before; else wastefully, we do fix operation again
				if(startpairskels.find(make_pair(ngbriter->origfstedges, blkiter->origfstedges)) != startpairskels.end())	//seen before
					continue;
				else
					startpairskels[make_pair(blkiter->origfstedges, ngbriter->origfstedges)] = 1;
				blockskel mergedskel = getMergedSkel(*blkiter, *ngbriter, adjlist, nodedismin, numnodes, &allblockskels);
				candmergeskels[mergedskel.origfstedges] = mergedskel; //now the cand blk is legal
			}
		}
		if(V) cout << "Number of candidate merges: " << candmergeskels.size() << endl;
		memocandskels = candmergeskels;
	
		//iterate through each candidate and obtain the legal block (paths/prob) for each
		//then reaplce its orig blocks with this, and recompute overallprob; among all such overallprobs, take the highest cand
		if(V) cout << "Computing blocks for the candidates\n";
		double highestoverallprob = 0.0;
		map < Graph, block> highestallblocksnew;		//the data structs for the chosen's candt (highest overallprob)
		vector <blockskel> highestallblockskelsnew;
		map <int, set <int> > highestadjlistnew;
		blockskel chosencandskel;		//the candskel that gives the highest overallprob - chosen to replace its parents!
		
		for(map < Graph, blockskel >::iterator iter = candmergeskels.begin(); iter != candmergeskels.end(); iter++) {
			block candblk = getBlockData(ifst, iter->second, numnodes, K, 1);	//1 means, we store paths/wts
			//this partclr blks's parent blks are replaced by it 
			map < Graph, block> allblocksnew;		//a list of all current blocks, each keyed by its origfstedges set
			vector <blockskel> allblockskelsnew;
			map <int, set <int> > adjlistnew;
			updateStructures(iter->second, candblk, &allblocks, &allblockskels, &adjlist, &allblocksnew, &allblockskelsnew, &adjlistnew);
			double overallprobnow = edgePushDynProb(&allblockskelsnew, &allblocksnew);
			memocandskelprobs[(iter->second).origfstedges] = overallprobnow;
			if(overallprobnow > highestoverallprob) {
				highestoverallprob = overallprobnow;
				highestallblockskelsnew = allblockskelsnew;
				highestallblocksnew = allblocksnew;
				highestadjlistnew = adjlistnew;
				chosencandskel = iter->second;
			}
		}//end for on candts

		allblocks = highestallblocksnew;
		allblockskels = highestallblockskelsnew;
		adjlist = highestadjlistnew;
		overallprob = highestoverallprob;
		if(V) cout << "After initial iteration, number of blocks: " << allblocks.size() << " and overall probability " << overallprob << endl;

		//now start loop; if new numblks is still > M
		//basically, three steps:
		//1 remove the chosen cand from memo
		//2 remove candts in memo have a partial overlap with chosen cand (if candt >= chosen cand, ie is a superset or equal, retain them)
		//3 get the ngbrs of chosen cand, and the corresp set of legal cand skels; unite with memoized set
		//now we have new cand skels set for this iteration -> repeat entire stuff, and get new chosen cand skel
		//input to this is the chosen cand skel and the new allblocks, blocksels and adjlist
		while (allblocks.size() > M) {
			if(V) cout << "Number of blocks: " << allblocks.size() << " > m " << M << endl;
			//step1
			int found = memocandskels.erase(chosencandskel.origfstedges);
			if(found == 0) {
				cout << "Error: Could not find chosen candidate in memo!\n";
				exit(1);
			}
			else {
				memocandskels.erase(chosencandskel.origfstedges);
				memocandskelprobs.erase(chosencandskel.origfstedges);
			}
			
			//step2
			vector < Graph > tobedeleted;		//list of keys into memo to be deleted due to partial overlap
			for(map < Graph, blockskel >::iterator miter = memocandskels.begin(); miter != memocandskels.end(); miter++) {
				blockskel thisskel = miter->second;
				//check the intersection; if it is not null and not equal to the chosen one, it means there is a partial overlaps and we need to delete it
				Graph intsxn = intersectSetsEdges(chosencandskel.origfstedges, thisskel.origfstedges);
				if((intsxn.size() > 0) && (intsxn != chosencandskel.origfstedges))
					tobedeleted.push_back(thisskel.origfstedges);
			}
			for(vector < Graph >::iterator diter = tobedeleted.begin(); diter != tobedeleted.end(); diter++) {
				map < Graph, blockskel >::iterator found = memocandskels.find(*diter);
				if(found == memocandskels.end()) {
					cout << "Error: Could not find candidate to be deleter in memo!\n";
					exit(1);
				}
				else {
					memocandskels.erase(*diter);
					memocandskelprobs.erase(*diter);
				}
			}
			
			//step3
			map < Graph, blockskel > newcandmergeskels;	//the new tobeadded bkls of the chosen cand skel
			map < pair < Graph, Graph >, int > startpairskels;	//for each cand, we start with a pair of legal skels	
			vector <blockskel> newngbrskels = getNgbrSkels(chosencandskel, &allblockskels);		//get list of nghbrng blkskels

			for(vector <blockskel>::iterator ngbriter = newngbrskels.begin(); ngbriter != newngbrskels.end(); ngbriter++) {	//for each ngbr we do for loop
				if(ngbriter->origfstedges == chosencandskel.origfstedges)
					continue;
				if(startpairskels.find(make_pair(ngbriter->origfstedges, chosencandskel.origfstedges)) != startpairskels.end()) //seen rev before
					continue;
				else
					startpairskels[make_pair(chosencandskel.origfstedges, ngbriter->origfstedges)] = 1;
				blockskel mergedskel = getMergedSkel(chosencandskel, *ngbriter, adjlist, nodedismin, numnodes, &allblockskels);
				newcandmergeskels[mergedskel.origfstedges] = mergedskel; //now the cand blk is legal
			}
			for(map < Graph, blockskel >::iterator aiter = newcandmergeskels.begin(); aiter != newcandmergeskels.end(); aiter++) //add these new candts to memo
				memocandskels[aiter->first] = aiter->second;
		
			//now get the best cand from the current memoized canddidate set; if the prob is present, we neednt compute it again
			if(V) cout << "Computing block probs of " << memocandskels.size() << " candidates\n";
			double newhighestoverallprob = 0.0;
			blockskel newchosencandskel;		//the candskel that gives the highest overallprob - chosen to replace its parents!
			for(map < Graph, blockskel >::iterator iter = memocandskels.begin(); iter != memocandskels.end(); iter++) {
				//check if blk exists in memo; if so, skip getting blk again
				map < Graph, double >::iterator found = memocandskelprobs.find(iter->first);
				if(found != memocandskelprobs.end()) {
					if((found->second) > newhighestoverallprob) {
						newhighestoverallprob = found->second;
						newchosencandskel = iter->second;
					}
				}
				else {	//this blk doenst have prob, need to compute it; this partclr blks's parent blks are replaced by it 
					block newcandblk = getBlockData(ifst, iter->second, numnodes, K, 1);
					map < Graph, block> newallblocksnew;		//a list of all current blocks, each keyed by its origfstedges set
					vector <blockskel> newallblockskelsnew;
					map <int, set <int> > newadjlistnew;
					updateStructures(iter->second, newcandblk, &allblocks, &allblockskels, &adjlist, &newallblocksnew, &newallblockskelsnew, &newadjlistnew); 
					double overallprobnow = edgePushDynProb(&newallblockskelsnew, &newallblocksnew);
					memocandskelprobs[(iter->second).origfstedges] = overallprobnow;
					if(overallprobnow > newhighestoverallprob) {
						newhighestoverallprob = overallprobnow;
						newchosencandskel = iter->second;
					}
				}//end else on whether prob is in memo
			}//end for on candts
			
			//get the blk of the new chosen cand; this partclr blks's parent blks are replaced by it 
			block newchosenblk = getBlockData(ifst, newchosencandskel, numnodes, K, 1);
			map < Graph, block> nextallblocks;		//a list of all current blocks, each keyed by its origfstedges set
			vector <blockskel> nextallblockskels;
			map <int, set <int> > nextadjlist;
			updateStructures(newchosencandskel, newchosenblk, &allblocks, &allblockskels, &adjlist, &nextallblocks, &nextallblockskels, &nextadjlist);
			
			//update the data structus for next iteration
			allblocks = nextallblocks;	
			allblockskels = nextallblockskels;
			adjlist = nextadjlist;
			overallprob = newhighestoverallprob;
			chosencandskel = newchosencandskel;
		}//end while numblks > M
	}//end else num blks > M

	foutstac << allblocks.size() << " " << overallprob << endl;
	if(V) cout << "Finally, number of blocks: " << allblocks.size() << " while m " << M << " and overall probability : " << overallprob << endl;
	map < Graph, int > blocknumbers;		//this is the block numbering to be used for stacgraph
	vector < Graph > blocknumbersrev;		//this is the block numbering to be used for path printing -> uniq blk nums in list order
	int startnum = 0;
	
	//print out the blockskel and blockprobs
	for(vector <blockskel>::iterator biter = allblockskels.begin(); biter != allblockskels.end(); biter++) {
		if(blocknumbers.find(biter->origfstedges) == blocknumbers.end()) {	//unseen skel
			blocknumbers[biter->origfstedges] = startnum;
			blocknumbersrev.push_back(biter->origfstedges);
			startnum++;
		}
	}
	//get the blkskelgraph after edge pushing and print out to stacfile in stac format! print the paths and wts
	map < Graph, vector <Graph > > blkskelgraph = getBlockSkelGraph(&allblockskels);
	for(map < Graph, vector < Graph > >::iterator iter = blkskelgraph.begin(); iter != blkskelgraph.end(); iter++) {
		foutstac << blocknumbers[iter->first] << "\t";
		for(vector < Graph >::iterator siter = (iter->second).begin(); siter != (iter->second).end(); siter++)
			foutstac << blocknumbers[*siter] << "\t";
		foutstac << endl;
	}
	for(int i=0; i < blocknumbersrev.size(); i++) {
		int numpths = allblocks[blocknumbersrev[i]].topkpaths.size();
		foutstac << i << " " << blocknumbersrev[i].size() << " " << numpths << endl;
		for(int k=0; k < numpths; k++)
			foutstac << allblocks[blocknumbersrev[i]].topkpaths[k] << "\n" << allblocks[blocknumbersrev[i]].topkpathwts[k] << endl;
	}
	return 0;
}
