/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "approximation/approximationUtils.h"
	
set <int> intersectSets(set <int> s1, set <int> s2) {
	set <int> ret;
	std::set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(), std::inserter(ret, ret.begin()));
return ret;
}

set < pair<int, int> > intersectSetsEdges(set < pair<int, int> > s1, set < pair<int, int> > s2) {
	Graph ret;
	std::set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(), std::inserter(ret, ret.begin()));
	return ret;
}
	
set <int> uniteSets(set <int> s1, set <int> s2) {
	set <int> ret;
	std::set_union(s1.begin(), s1.end(), s2.begin(), s2.end(), std::inserter(ret, ret.begin()));
	return ret;
}

set < pair<int, int> > uniteSetsEdges(set < pair<int, int> > s1, set < pair<int, int> > s2) {
	set < pair<int, int> > ret;
	std::set_union(s1.begin(), s1.end(), s2.begin(), s2.end(), std::inserter(ret, ret.begin()));
	return ret;
}

//subset check - is s1 a subset (or equal to) of s2
int subsetCheck(set <int> s1, set <int> s2) {
	set <int> inter = intersectSets(s1, s2);
	return (inter == s1 ? 1 : -1);
}

int subsetCheckEdges(set < pair<int, int> > s1, set < pair<int, int> > s2) {
	set < pair<int, int> > inter = intersectSetsEdges(s1, s2);
	return (inter == s1 ? 1 : -1);
}

//get the ngbrng skels of the given skel (these are legal skels)
vector <blockskel> getNgbrSkels(blockskel thisskel, vector <blockskel>* allskels) {
	vector <blockskel> ret;
	//iterate thro all blks and see which ones share a min/max with this min/max
	for(vector <blockskel>::iterator iter = allskels->begin(); iter != allskels->end(); iter++)
		if((iter->minnode == thisskel.minnode) || (iter->minnode == thisskel.maxnode) || (iter->maxnode == thisskel.maxnode) || (iter->maxnode == thisskel.minnode))
			ret.push_back(*iter);
	return ret;
}
	
//get the top K paths and probs for the given subfst block -> a legal block; so we can iterate the origfstnodes set to build statepr
block getBlockData(StdVectorFst *ifst, blockskel thisskel, int numnodes, int K, int storepaths) {
	block retblk;
	retblk.minnode = thisskel.minnode;
	retblk.maxnode = thisskel.maxnode;
	int statepr0[numnodes];			//bitmap of states present in this subfst
	for(int s=0; s<numnodes; s++)
		statepr0[s] = (thisskel.origfstnodes.find(s) != thisskel.origfstnodes.end()) ? 1 : 0;

	StdVectorFst fstblock0 = projectSubFst_edges(ifst, numnodes, thisskel.minnode, thisskel.maxnode, statepr0, &thisskel.origfstedges);
	PathIterator incrementalN;
	constructIterator(&incrementalN, &fstblock0);
	for(int j=0;j < K; j++)	{
		Path p = incrementalN.getNext();
		if(incrementalN.moreflag == 0)	//no more paths
			break;
		double cost = p.getPathCost();
		retblk.totprob += exp(-1.0*cost);	//prob is e^-wt
		if(storepaths == 1) {
			string path = p.getPathStr();
			retblk.topkpathwts.push_back(cost);
			retblk.topkpaths.push_back(path);
		}
	}
	delete incrementalN.enumerator;
	return retblk;
}

//check if there are any middle-intervening blocks in the given block - which has a unique max and min node; allbocks list contains only legal blocks
void fixMidlSkel(blockskel *thisskel, vector <blockskel> *allskels) {
	set <int> newset;	//new set of orignodes
	Graph newsetedges;	//new set of origedges
	for(vector <blockskel>::iterator iter = allskels->begin(); iter != allskels->end(); iter++) {
		//if its min node lies in thisskel, but does not equal thisskel's min/max, then assimilate it
		if((iter->minnode != thisskel->minnode) && (iter->minnode != thisskel->maxnode) && ((thisskel->origfstnodes).find(iter->minnode) != (thisskel->origfstnodes).end())) {
			newset = uniteSets(newset,  iter->origfstnodes);
			newsetedges = uniteSetsEdges(newsetedges, iter->origfstedges);
			thisskel->parskels.insert(iter->origfstedges);
		}
		//similarly for its maxnode
		else if((iter->maxnode != thisskel->minnode) && (iter->maxnode != thisskel->maxnode) && ((thisskel->origfstnodes).find(iter->maxnode) != (thisskel->origfstnodes).end())) {
			newset = uniteSets(newset,  iter->origfstnodes);
			newsetedges = uniteSetsEdges(newsetedges, iter->origfstedges);
			thisskel->parskels.insert(iter->origfstedges);
		}
	}
	thisskel->origfstnodes = uniteSets(newset, thisskel->origfstnodes);		//add the extra nodes to the set
	thisskel->origfstedges = uniteSetsEdges(newsetedges, thisskel->origfstedges);		//add the extra edges to the set
}

//checking if there is a uniq maxnode/minnode in the given skel; if yes, set maxnode/minnode; else, Fix it, and then set them accordingly
//for the Fix, this obtains the minimum node to which there is path from all end pts of mergedskel
//corresp for minnode, we use adjlistrev, and Fix obtain max node from which there is path to all start pts of mergedskel
//the minmaxflag is -1 if we want to do fixmin, and 1 if we want to do fixmax
void fixMinMaxSkel(blockskel *thisskel, map <int, set<int> > adjlist, int* nodedismin, int numnodes, vector <blockskel> *allblockskels, int minmax) {
	//if any of the skel nodes have outvxs set of size 0, it is _a_ maxnode
	//the out vals would corresp to in vals, if adjlist corresp to reversed edges in orig fst
	map <int, int> inoutdeg;	//gives the outdeg of nodes in skel; if 0, it is a maxnode; used as counterpart for indeg for fixmin
	int inout0node;
	int numinout0 = 0;
	for(set <int>::iterator iter = (thisskel->origfstnodes).begin(); iter != (thisskel->origfstnodes).end(); iter++) {
		inoutdeg[*iter] = 0;
		for(Graph::iterator siter = (thisskel->origfstedges).begin(); siter != (thisskel->origfstedges).end(); siter++) {
			if(((minmax == 1)&&((*iter) == (siter->first))) || (minmax == -1)&&((*iter) == (siter->second)))	//an edge leaving iter in the block
				inoutdeg[*iter] += 1;
		}
		if(inoutdeg[*iter] == 0) {
			numinout0 += 1;
			inout0node = *iter;
		}
	}
	if(numinout0 == 1) {
		if(minmax == 1)
			thisskel->maxnode = inout0node;
		else
			thisskel->minnode = inout0node;
	}
	else if(numinout0 < 1) {
		cout << "Error: fixMinMaxSkel: No min/max node!\n";
		exit(1);
	}
	else {	//more than one maxnode; go thro list of all nodes in origfst, and see which ones are reachable from all maxnodes of skel, then get min among those
		set <int> meetpts;
		for(map <int, set<int> >::iterator iter = adjlist.begin(); iter != adjlist.end(); iter++) {
			int rchblflag = 1;		//alter the rchblty info in the skel for loop, by anding for all nodes in endset
			for(set <int>::iterator siter = (thisskel->origfstnodes).begin(); siter != (thisskel->origfstnodes).end(); siter++) {
				if(inoutdeg[*siter] != 0) //siter has to be an end node!
					continue;
				bool pamax = (nodedismin[(*siter) * numnodes + (iter->first)] == INFIN); //no path to this iter from this siter
				bool pamin = (nodedismin[(iter->first) * numnodes + (*siter)] == INFIN); //no path from iter to this endnnode
				if(((minmax == 1) && pamax) || ((minmax == -1) && pamin))
					rchblflag = 0;
			}
			if(rchblflag == 1) //this origfstnode is a meetpt
				meetpts.insert(iter->first);
		}
		if(meetpts.size() == 0) {
			cout << "Error: fixMinMaxSkel: No meet points!\n";
			exit(1);
		}
		int extrmeetpt = *(meetpts.begin()); //now, we get the min meet pt
		for(set <int>::iterator iter = meetpts.begin(); iter != meetpts.end(); iter++) {
			if(((minmax == 1)&&(extrmeetpt > (*iter))) || ((minmax == -1)&&(extrmeetpt < (*iter))))
				extrmeetpt = *iter;
		}
		//for the parent lineage thing, keep tracck of the set of nodes being added here, incl meet pt and end nodes
		//then, check which current skels are a subset of this lineage set, and add them as parent of thisskel
		set <int> lineageset;
		lineageset.insert(extrmeetpt);		//adding the meeting node
		(thisskel->origfstnodes).insert(extrmeetpt);
		if(minmax == 1)
			thisskel->maxnode = extrmeetpt;
		else
			thisskel->minnode = extrmeetpt;

		//now we have the extrmeetpt, and we can do the fix -> basically, add all origfstnodes that lie on a path from any of the end nodes and the extrmeetpt
		for(map <int, set<int> >::iterator iter = adjlist.begin(); iter != adjlist.end(); iter++) {
			for(set <int>::iterator siter = (thisskel->origfstnodes).begin(); siter != (thisskel->origfstnodes).end(); siter++) {
				if(inoutdeg[*siter] != 0)
					continue;
				lineageset.insert(*siter);	//adding the end pt
				bool pathmax = ((nodedismin[(*siter) * numnodes + (iter->first)] != INFIN) && (nodedismin[(iter->first) * numnodes + extrmeetpt] != INFIN) && (extrmeetpt != (*siter)) && (extrmeetpt != (iter->first))); //path from siter to iter to extrmeetpt in case of fixMax
				bool pathmin = ((nodedismin[(iter->first) * numnodes + (*siter)] != INFIN) && (nodedismin[(extrmeetpt) * numnodes + iter->first] != INFIN) && (extrmeetpt != (*siter)) && (extrmeetpt != (iter->first))); //path from extrmeept to iter to siter in case of fixMin
				if( ((minmax == 1) && pathmax) ||	((minmax == -1) && pathmin)) {
					(thisskel->origfstnodes).insert(iter->first);	//add this node to the skel
					lineageset.insert(iter->first);	//adding the internal node
				}
			}
		}
		//the key difference is now; the parent skels are determined by the subsetting on the edgeset, not the nodeset; also need to assiml lineage edges
		Graph lineagesetedges;	//the set of edges among any pair of nodes in the lineage nodes set; reprsnts all paths to meetpt
		for(set <int>::iterator iter = lineageset.begin(); iter != lineageset.end(); iter++) {
			for(set <int>::iterator siter = lineageset.begin(); siter != lineageset.end(); siter++) {
				if(adjlist[*iter].find(*siter) != adjlist[*iter].end())	//iter to siter edge exists
					lineagesetedges.insert(make_pair(*iter, *siter));
				else if (adjlist[*siter].find(*iter) != adjlist[*siter].end()) //siter to iter
					lineagesetedges.insert(make_pair(*siter, *iter));
			}
		}
		thisskel->origfstedges = uniteSetsEdges(thisskel->origfstedges, lineagesetedges); //assiml lineage edges to origedges, determine parents
		for(vector <blockskel>::iterator iter = allblockskels->begin(); iter != allblockskels->end(); iter++) {
			int issub = subsetCheckEdges(iter->origfstedges, lineagesetedges);
			if(issub == 1)
				thisskel->parskels.insert(iter->origfstedges);		//this parblk has been subsumed
		}
	}//end else, for fix opertn
}

//obtain overallprob of the current block structure (all blks are legal); we also memoize the prob of a partclr block, so as not to recompute it
double probMultDyn(map < Graph, vector < Graph > >  *blkskelgraph, map < Graph, block> *allblocks, Graph startskelset, map < Graph, double> *memo) {
	if(memo->find(startskelset) != memo->end())	//we have memoized this
		return (*memo)[startskelset];
	double curprob = (startskelset.find(make_pair(-1, 0)) != startskelset.end()) ? 1 : (*allblocks)[startskelset].totprob;	// dummy set
	//if no children, it is current blk's totprob; else if children, it is cur totprob * sum of children's totprobfunc (recursive)
	if((*blkskelgraph)[startskelset].size() == 0) {	//no children
		(*memo)[startskelset] = curprob;	//memoize it
		return curprob;
	}
	else {		//obtain sum of children's totprobfunc
		double sumchildren = 0.0;
		for(vector < Graph >::iterator iter = (*blkskelgraph)[startskelset].begin(); iter != (*blkskelgraph)[startskelset].end(); iter++)
			sumchildren += probMultDyn(blkskelgraph, allblocks, *iter, memo);
		(*memo)[startskelset] = curprob * sumchildren;	//memoize it
		return (curprob * sumchildren);
	}
}


//this functions takes in the list of blockskels and computes the blockskels adj list, based on min/max nodes of each block (they are all legal)
map < Graph, vector <Graph > > getBlockSkelGraph(vector <blockskel> *skels) {
	map < Graph, vector <Graph > > retgph;
	for(vector <blockskel>::iterator iter = skels->begin(); iter != skels->end(); iter++) {
		vector <Graph > emptyset;
		retgph[iter->origfstedges] = emptyset;		//need to initlze all skels, otherwise the outfdeg0 skels dont show up in graph
		for(vector <blockskel>::iterator iter2 = skels->begin(); iter2 != skels->end(); iter2++) {
			if(iter->maxnode == iter2->minnode) {	//ie, we can go to blk2 from blk1 (they are legal blocks)
				retgph[iter->origfstedges].push_back(iter2->origfstedges);
			}
		}	
	}
	return retgph;
}

//go thro the allskels list and see which ones are subsets of thisset, return those
set <Graph > getParentSkels(Graph thisset, vector <blockskel> *allblockskels) {
	set < Graph > retset;
	for(vector <blockskel>::iterator iter = allblockskels->begin(); iter != allblockskels->end(); iter++) {
		int issubset = subsetCheckEdges(iter->origfstedges, thisset);
		if(issubset == 1)
			retset.insert(iter->origfstedges);
	}
	return retset;
}

//this function takes in the list of blockskels and their corresp blocks, and applies edge pushing, and then compute overall prob by dyn progmg
double edgePushDynProb(vector <blockskel> *allblockskels, map < Graph, block> *allblocks) {
	//do edgepushing and obtain the graph of blockskels, and compute the entire prob by dyn prog
	map < Graph, vector <Graph > > blkskelgraph = getBlockSkelGraph(allblockskels);
	//check if the blkskels have a uniq start skel (ie minnode is 0 -> fststartstate for exactly 1 skel; if not, add 'dummy skel', and make it start skel)
	vector < Graph > min0skelsets;
	Graph startskelset;
	for(vector <blockskel>::iterator iter = allblockskels->begin(); iter != allblockskels->end(); iter++) {
		if(iter->minnode == 0) {
			min0skelsets.push_back(iter->origfstedges);
			startskelset = iter->origfstedges;
		}
	}
	if(min0skelsets.size() == 0) {
		cout << "Error: edgePushDynProb: No start blocks!\n";
		exit(1);
	}
	else if(min0skelsets.size() > 1) {	//need to insert a dummy skel (origset is simply -1, 0 (maxnode) and totprob is 1 used later) into the blkskelgraph
		Graph dummyskelset;
		dummyskelset.insert(make_pair(-1,0));	//dummy blk -1 to 0
		for(vector < Graph >::iterator iter = min0skelsets.begin(); iter != min0skelsets.end(); iter++) //add to blkskelgrpah and set outbkls to those in min0skelsets
			blkskelgraph[dummyskelset].push_back(*iter);
		startskelset = dummyskelset;
	}
	map < Graph, double > memo;
	return probMultDyn(&blkskelgraph, allblocks, startskelset, &memo);
}

//compute new adjlist based on the new skels; here, we 'collapse' a block(skel) to a new 'super-edge' in the graph, and include only min/max nodes of it in the graph, as the super-edge's end pts
map <int, set <int> > adjlistFromSkels(vector <blockskel> *allblockskelsnew) {
	map <int, set <int> > retmap;
	//iterate thro skels, and simply put edges between block's min and max nodes; rest are autom taken care of; also get largest maxnode, and put an empty set for that
	int largestmaxnode = 0;
	for(vector <blockskel>::iterator biter = allblockskelsnew->begin(); biter != allblockskelsnew->end(); biter++) {
		retmap[biter->minnode].insert(biter->maxnode);
		if((biter->maxnode) > largestmaxnode)
			largestmaxnode = biter->maxnode;
	}
	set <int> emptyset;
	retmap[largestmaxnode] = emptyset;
	return retmap;
}

//given the blockskel and old structs, update and return new structs
void updateStructures(blockskel thisblkskel, block thisblk, map < Graph, block> *allblocks, vector <blockskel> *allblockskels, map <int, set <int> > *adjlist, map < Graph, block> *allblocksnew, vector <blockskel> *allblockskelsnew, map <int, set <int> > *adjlistnew) { 
	//compute allblockskelsnew and allblocksnew
	set < Graph > parentskels = getParentSkels(thisblkskel.origfstedges, allblockskels);
	for(vector <blockskel>::iterator biter = allblockskels->begin(); biter != allblockskels->end(); biter++) {
		if(parentskels.find(biter->origfstedges) == parentskels.end()) { //a skel that isnt a parent of this cand blk
			if((*allblocksnew).find(biter->origfstedges) == (*allblocksnew).end()) {	//this wasnt seen before
				(*allblockskelsnew).push_back(*biter);
				(*allblocksnew)[biter->origfstedges] = (*allblocks)[biter->origfstedges];				
			}
		}
	}			
	(*allblockskelsnew).push_back(thisblkskel);	//adding this new blk's skel
	(*allblocksnew)[thisblkskel.origfstedges] = thisblk;
	*adjlistnew = adjlistFromSkels(allblockskelsnew);
}


blockskel getMergedSkel(blockskel chosencandskel, blockskel ngbr, map <int, set<int> > adjlist, int* nodedismin, int numnodes, vector <blockskel> *allblockskels) {
	blockskel mergedskel;		//this is a would-be candidtate
	mergedskel.origfstnodes = uniteSets(chosencandskel.origfstnodes, ngbr.origfstnodes);	//unite the current ngbr with the cur blk
	mergedskel.origfstedges = uniteSetsEdges(chosencandskel.origfstedges, ngbr.origfstedges);	//unite the current ngbr with the cur blk
	mergedskel.parskels.insert(chosencandskel.origfstedges);
	mergedskel.parskels.insert(ngbr.origfstedges);
	while(1) {	//keep iterating and fixing as long as the cand blk is illegal
		blockskel copyskel = mergedskel;	//to be used at end of itern to check if there has been any change
		fixMinMaxSkel(&mergedskel, adjlist, nodedismin, numnodes, allblockskels, 1);		//fix maxnode if needed
		fixMinMaxSkel(&mergedskel, adjlist, nodedismin, numnodes, allblockskels, -1);		//fix minnode if needed
		fixMidlSkel(&mergedskel, allblockskels);
		if (mergedskel.origfstedges.size() == copyskel.origfstedges.size()) //no more cahnges - its now a legal candidate
			break;
	}
	return mergedskel;
}
