/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "frontends/SQLUtils.h"

int unpack_int32(char *data) {
  return ntohl(*((uint32_t*) data));
}

double unpack_double(char* data) {
  union {
    double d;
    unsigned int n[2];
  } swap;
  memcpy(&(swap.n), data, sizeof(unsigned int)*2);
  unsigned int tmp = ntohl(swap.n[0]);
  swap.n[0]        = ntohl(swap.n[1]);
  swap.n[1]        = tmp;
  return swap.d;
}

void printUsage (SQLType sqltype) {
	string always = "<db name> <port number> <SQL query and regex: \"select <aux cols>, exists <data col> in_top <num ans> from <table> where <data col> like\" \"<regex>\"> ";
	string verbose = "-v (optional, verbose)\n";
	switch(sqltype) {
		case KMAP:
			cout << "Usage: kmapSQL " << always << "-k <k value> " << verbose;
			break;
		case STAC:
			cout << "Usage: staccatoSQL " << always << "-k <k value> -m <m value> " << verbose;
			break;
		case STACPROJ:
			cout << "Usage: staccatoSQLIndex " << always << "-k <k value> -m <m value> -i <index term> -l <(optional) length of fixed-length regex> " << verbose;
			break;
		case FULLFST:
			cout << "Usage: fullFSTSQL " << always << verbose;
			break;
	}
}

int dfacheck (string inpstr, int* dfaquery, int inpstate, int NUMST, int startstate, int finalstate) {
	if(inpstate == finalstate)
		return finalstate;

	int len = inpstr.length();
	int curstate = inpstate;
	int nxtstate;
	for(int s=0; s<len; s++) {
		//get the next state frmo the query - check if char is ASCII (32-125) or not; if not, resend to startstate
		int inpstrsascii = inpstr[s] - '\0';
		if((inpstrsascii >= 32) && (inpstrsascii < 126)) {	//alphabet of the fst
			nxtstate = dfaquery[curstate*94 + (inpstr[s] - ' ')];
			if (nxtstate == finalstate)
				return finalstate;
		}
		else
			nxtstate = startstate;
		curstate = nxtstate;	//move to the next state as per dfa
	}
	if(len == 0) {
		if(curstate != finalstate)
			curstate = startstate;		//reset dfa to start (assumed 0)
		else
			curstate = finalstate;		//stays in finalstate
	}
	return curstate; //return the curstate reached after having read the full string
}

void sortResults (vector <restuple> *results) {
	int rmapsize = results->size();
	restuple tmp;
	for(int i=0; i<rmapsize-1; i++) {
		for(int j=i+1; j<rmapsize; j++) {
			if((*results)[i].matchwt > (*results)[j].matchwt) {
				restuple tmp = (*results)[i];
				(*results)[i] = (*results)[j];
				(*results)[j] = tmp;
			}
		}
	}
}

void printResults (vector <restuple> results, vector <string> auxcolnames, int numAns) {
	int rmapsize = results.size();
	for(int au=0;au<auxcolnames.size(); au++) {
		cout << auxcolnames[au] << "\t\t";
	}
	cout << "Probability (-log weight)\n";
	for(int rr=0; (rr<rmapsize)&&(rr<numAns); rr++) {
		for(int oc=0; oc<auxcolnames.size(); oc++)
			cout << results[rr].precolvals[oc] << "\t";
		cout << exp(-1.0*results[rr].matchwt) << " (" << results[rr].matchwt << ")\n";
	}
}

querystructs parseQuery(char* sqlquery) {
	querystructs qstructs;
	char* tok = strtok(sqlquery, " ,");
	vector <string> toks;
	while(tok != NULL) {
		toks.push_back(tok);
		tok = strtok(NULL, " ,;'");
	}
	int ntok = 0;
	while(ntok < toks.size()) {
		if(ntok < toks.size() && toks[ntok].compare("select") == 0) {
			ntok++;
			while(toks[ntok].compare("exists") != 0) {
				qstructs.auxcolnames.push_back(toks[ntok]);
				ntok++;
			}
		}
		if((ntok < toks.size()) && (toks[ntok].compare("exists") == 0)) {
			qstructs.fstcolname = toks[ntok+1];
			qstructs.type = 1;
			ntok += 2;
			if(toks[ntok].compare("in_top") == 0) {
				qstructs.Ktop = atoi(toks[ntok+1].c_str());
				ntok += 2;
			}
			else //if there is no in_top clause, default value for K is 10
				qstructs.Ktop = 10;
		}
		if(ntok < toks.size() && toks[ntok].compare("from") == 0) {
			qstructs.tablename = toks[ntok+1];
			ntok += 2;
		}
		if(ntok < toks.size() && toks[ntok].compare("where") == 0) {
			if((toks[ntok+1].compare(qstructs.fstcolname) != 0)||(toks[ntok+2].compare("like") != 0)) {
				cout << "Error: Inadmissible 'where' clause!\n";
				exit(1);
			}
			ntok += 4;
		}
		ntok++;
	}
	return qstructs;
}
