/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "libraries/PathEnumerator.h"

PathEnumerator::~PathEnumerator() {
	int i;
	for (i = 0; i < graph_data_.size(); i ++)
		delete graph_data_[i];
	for (i = 0; i < graph_data_2_.size(); i ++)
		delete graph_data_2_[i];
	for (i = 0; i < edge_buffer_.size(); i ++)
		delete edge_buffer_[i];
}

// construct graph_data_
void PathEnumerator::fstToGraph() {
	int num_of_nodes = 0;
	int src_idx, sink_idx;
	int i, j;
	// get total number of nodes; indices of src and sink
	for (StateIterator< Fst<StdArc> > siter(fst_); !siter.Done(); siter.Next()) {
		ArcIterator< Fst<StdArc> > aiter(fst_, siter.Value());
		if (aiter.Done())
			sink_idx = num_of_nodes;
		else if (siter.Value() == fst_.Start())
			src_idx = num_of_nodes;
		assert(num_of_nodes == siter.Value());
		num_of_nodes ++;
	}
	// resize graph_data_ to allocate nodes
	graph_data_.resize(num_of_nodes);
	graph_data_2_.resize(num_of_nodes);
	for (i = 0 ; i < num_of_nodes; i ++) {
		graph_data_[i] = new Node();
		graph_data_[i]->index = i;
		graph_data_2_[i] = new Node();
		graph_data_2_[i]->index = i;
	}
	source_node_ = graph_data_[src_idx];
	sink_node_ = graph_data_[sink_idx];
	// iterate all arcs; add to in and out
	for (StateIterator< Fst<StdArc> > siter(fst_); !siter.Done(); siter.Next()) {
		int from = siter.Value();
		set<EdgeWeight, CompareEdgeWeight> unique_edge;
		unique_edge.clear();
		for(ArcIterator< Fst<StdArc> > aiter(fst_, from); !aiter.Done(); aiter.Next()) {
			Arc arc = aiter.Value();
			int to = arc.nextstate;
			double weight = arc.weight.Value();
			char label = arc.olabel;
			int of = arc.origfrom;
			int ot = arc.origto;
			edge_buffer_.push_back(new EdgeWeight(from, to, weight, label, of, ot));
			EdgeWeight *curr_edge = *(edge_buffer_.rbegin());
			if (unique_edge.find(*curr_edge) != unique_edge.end()) // check if it is already constructed
				continue;
			unique_edge.insert(*curr_edge); // for future checking
			graph_data_[from]->out.push_back(curr_edge); // add to out
			graph_data_[to]->in.push_back(curr_edge); // add to in
			if (graph_data_2_[to]->in.empty()) // weight_matrix
				graph_data_2_[to]->in.push_back(curr_edge);
			else {
				bool found = false;
				for (i = 0; i < graph_data_2_[to]->in.size(); i ++) {
					if ((graph_data_2_[to]->in[i]->from == curr_edge->from)	&& (graph_data_2_[to]->in[i]->weight > curr_edge->weight)) {
						graph_data_2_[to]->in[i] = curr_edge;
						found = true;
					}
				}
				if (!found)
					graph_data_2_[to]->in.push_back(curr_edge);
			}
		}//end arc iterator
	}//end state iterator
}

bool PathEnumerator::getNextBest(Path& resPath) {
	//check if first time called; if yes, do dijkstra with all edge reversed, build the shortest path table, return the entry of source; if no, pop min from the heap
	if (is_first_time_) {
		is_first_time_ = false;
		initShortestToSink();
		paths_heap_.push(shortest_to_sink_[source_node_->index]); // construction finished, push to the heap
	}
	// pop min from the heap, set as A[k]
	if (paths_heap_.empty())
		return false;
	A_.push_back(paths_heap_.top());
	paths_heap_.pop();
	// find deviation edge from A[k]
	int devEdgeNum = 0; // [0, len)
	set<EdgeWeight *> usedBestEdge_set; // edges that have been considered
	int i, j;
	usedBestEdge_set.clear();
	for (j = 0; j < k_; j ++) {
		// if A[j] and A[k] have same prefix from 0 to (i - 1)th edges, devEdgeNum should set to i
		for (i = 0; i < A_[k_].numOfEdges() && i < A_[j].numOfEdges(); i ++)
			if (A_[k_].getEdge(i) != A_[j].getEdge(i)) // check if the ith edge is the same
				break;
		assert(i < A_[k_].numOfEdges()); // they should not be the same
		assert(i < A_[j].numOfEdges());
		if (devEdgeNum < i) { // process every i; found a longer common path; update deviation edge to be i; and rebuild the set
			devEdgeNum = i;
			usedBestEdge_set.clear();
			usedBestEdge_set.insert(A_[j].getEdge(i));
		}
		else if (devEdgeNum == i)
			usedBestEdge_set.insert(A_[j].getEdge(i)); // add one more edge to the set
	}

	// expand one path (s=>u->v=>t) to be (s=>u->v'=>t) of u->v as the deviation edge until the last edge of A[k], with the min value of w_uv' + shortestLen_v't
	double shortest_from_u = INF;
	EdgeWeight *u_to_v_prime;
	bool found_v_prime = false;
	// remember to check if v' has been considered by A[j], 0 <= j < k, of course also A[k]; decide v', pv_prime means possible v'
	Node *u = graph_data_[A_[k_].getEdge(devEdgeNum)->from];
	for (j = 0; j < u->out.size(); j ++) {
		EdgeWeight *u_to_pv_prime = u->out[j]; // get next possible v
		if (usedBestEdge_set.find(u_to_pv_prime) != usedBestEdge_set.end()) // compute the cost funtion only if not used
			continue; // considered by A[j], j < k
		if (u_to_pv_prime == A_[k_].getEdge(devEdgeNum)) // considered by A[k]
			continue;
		// compute w_uv' + shortestLen_v't and choose the best one
		if (u_to_pv_prime->weight + shortest_to_sink_[u_to_pv_prime->to].getPathCost() < shortest_from_u) {
			shortest_from_u = u_to_pv_prime->weight + shortest_to_sink_[u_to_pv_prime->to].getPathCost();
			u_to_v_prime = u_to_pv_prime;
			found_v_prime = true;
		}
	}
	if (found_v_prime)
		addPathToHeap(A_[k_], devEdgeNum, u_to_v_prime); // construct s=>u->v'=>t and put it into the heap

	// process remaining nodes as u similarly, but not check the set
	for (i = devEdgeNum + 1; i < A_[k_].numOfEdges(); i ++) {
		double shortest_from_u = INF;
		EdgeWeight *u_to_v_prime;
		bool found_v_prime = false;
		// remember to check if v' has been considered by A[j], 0 <= j < k, of course also A[k]; decide v', pv_prime means possible v'
		Node *u = graph_data_[A_[k_].getEdge(i)->from];
		for (j = 0; j < u->out.size(); j ++) {
			// get next possible v
			EdgeWeight *u_to_pv_prime = u->out[j];
			if (usedBestEdge_set.find(u_to_pv_prime) != usedBestEdge_set.end()) // compute the cost funtion only if not used
				continue; // considered by A[j], j < k
			if (u_to_pv_prime == A_[k_].getEdge(i)) // considered by A[k]
				continue;
			// compute w_uv' + shortestLen_v't and choose the best one
			if (u_to_pv_prime->weight	+ shortest_to_sink_[u_to_pv_prime->to].getPathCost() < shortest_from_u) {
				shortest_from_u = u_to_pv_prime->weight + shortest_to_sink_[u_to_pv_prime->to].getPathCost();
				u_to_v_prime = u_to_pv_prime;
				found_v_prime = true;
			}
		}
		if (found_v_prime)
			addPathToHeap(A_[k_], i, u_to_v_prime); // construct s=>u->v'=>t and put it into the heap
	}
	resPath = A_[k_];
	k_ ++;
	return true;
}

bool PathEnumerator::addPathToHeap(Path &A_k, int index_of_uv, EdgeWeight *u_to_v_prime) {
	Path heap_path;
	heap_path.appendPath(A_k, 0, index_of_uv); // s=>u
	heap_path.appendEdge(u_to_v_prime); // u->v
	heap_path.appendPath(shortest_to_sink_[u_to_v_prime->to]); // v=>t
	paths_heap_.push(heap_path); // construction finished, push to the heap
	return true;
}

void PathEnumerator::initShortestToSink() {
	// perform dijkstra from sink
	int GRAPH_NODE_NUM = graph_data_.size(); //= sink_node_->index + 1; we get the size explicitly
	Dist* dist = new Dist[GRAPH_NODE_NUM];
	bool* visited = new bool[GRAPH_NODE_NUM];
	int i;
	for (i = 0; i < GRAPH_NODE_NUM; i ++) {
		visited[i] = false;
		dist[i].index = i;
		dist[i].cost = INF;
		dist[i].edge = NULL;
	}
	dist[sink_node_->index].cost = 0;
	priority_queue<Dist, vector<Dist>, CompareDist> heap;
	heap.push(dist[sink_node_->index]);
	for (i = 0; i < GRAPH_NODE_NUM; i ++) {
		bool found = false;
		Dist d;
		while(!heap.empty()) {
			d = heap.top();
			heap.pop();
			if (!visited[d.index]) {
				found = true;
				break;
			}
		}
		if (!found)
			break;
		int v = d.index;
		visited[v] = true;
		int j;
		for (j = 0; j < graph_data_2_[v]->in.size(); j ++) {
			EdgeWeight *curr_edge = graph_data_2_[v]->in[j];
			if (dist[curr_edge->from].cost  > dist[v].cost + curr_edge->weight && visited[curr_edge->from] == false) {
				dist[curr_edge->from].cost = dist[v].cost + curr_edge->weight;
				dist[curr_edge->from].edge = curr_edge;
				heap.push(dist[curr_edge->from]);
			}
		}
	}
	// fill in shortestToSink
	shortest_to_sink_.resize(GRAPH_NODE_NUM);
  for (i = 0; i < GRAPH_NODE_NUM; i ++) {
		int index = dist[i].index;
		if(visited[index] == false)
			continue;
		while (index != sink_node_->index) {
			shortest_to_sink_[i].appendEdge(dist[index].edge);
			index = dist[index].edge->to;
		}
	}
	delete []visited;
	delete []dist;
}

Dist& Dist::operator=(const Dist& rhs) {
	this->index = rhs.index;
	this->cost = rhs.cost;
	this->edge = rhs.edge;
}
